/*
 * Si5351C 10MHz-to-27MHz frequency tanslator for Ultimate 3S WSPR Beacon
 * File:   main.c
 * Author: Bert, VE2ZAZ - http://ve2zaz.net
 * Created on February 28, 2018
 * Target: Microchip PIC12F683
 * Environment: MPLABX IDE v5.30, XC8 Compiler v1.45
 *
 * The following parameters are used to convert a 10.000 MHz REFIN frequency to 27.000 MHz output frequency
 * 
 * 10 MHz x 81
 * -----------  = 27 MHz
 *     30
 * 
 * MSNA
 * ----
 * a=81, b=0, c=1  ->   MSNA_P1 = (128 x 81) - 512 = 9856 = 0x2680
 *                      MSNA_P2 = 0
 *                      MSNA_P3 = 1

 * 
 * MS0
 * ---
 * a=30, b=0, c=1  ->   MS0_P1 = (128 x 30) - 512 = 3328 = 0x0D00
 *                      MS0_P2 = 0
 *                      MS0_P3 = 1
 *
 * R0 = 1,  FBA_INT = 0
 *
*/

// COMPILER DIRECTIVES
#include <xc.h>
#include <pic12f683.h>
#define  test_bit(var, bit)   ((var) & (1 <<(bit)))
#define  set_bit(var, bit)    ((var) |= (1 << (bit)))
#define  clear_bit(var, bit)    ((var) &= ~(1 << (bit)))
#define _XTAL_FREQ                  8000000

// GLOBALS
unsigned char		i;
signed char         j;
unsigned char       temp_buffer;     
unsigned char       status;
unsigned char       old_status;


// CONFIGURATION BYTES
#pragma config FOSC = INTOSCIO  // Oscillator Selection bits (INTOSCIO oscillator: I/O function on RA4/OSC2/CLKOUT pin, I/O function on RA5/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select bit (MCLR pin function is digital input, MCLR internally tied to VDD)
#pragma config CP = OFF         // Code Protection bit (Program memory code protection is disabled)
#pragma config CPD = OFF        // Data Code Protection bit (Data memory code protection is disabled)
#pragma config BOREN = OFF      // Brown Out Detect (BOR disabled)
#pragma config IESO = OFF       // Internal External Switchover bit (Internal External Switchover mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enabled bit (Fail-Safe Clock Monitor is disabled)

// CONSTANTS AND MASKS DEFINITION

	//Other definitions
#define scl 				4       // The bit position on GPIO 
#define sda 				5       // The bit position on GPIO 
#define LED 				2       // The bit position on GPIO for the LED
#define i2c_addr            0x60    // The Si5351 I2C address


// Timer routines

void delay_ms(int delay) {
    for (i = 0; i < delay; i++) __delay_ms(1);
}

void delay_100us(int delay) {
    for (i = 0; i < delay; i++) __delay_us(100);
}

void i2c_send_start(void)
{
    clear_bit(GPIO, sda);
    delay_100us(1);
    clear_bit(TRISIO, sda);             // Sets the line to output to capture it
    delay_100us(1);
    clear_bit(GPIO, scl);
    delay_100us(1);
}
    
void i2c_send_stop(void)
{
    clear_bit(GPIO, sda);             
    delay_100us(1);
    set_bit(GPIO, scl);
    delay_100us(1);
    set_bit(GPIO, sda);             
    delay_100us(1);
    set_bit(TRISIO, sda);             // Sets the line to input to release it
    delay_100us(1);
}

void i2c_send_addr(char buffer)
{
     for (j=6;j>=0;j--)
    {
        if (test_bit(buffer,j)) set_bit(GPIO, sda);         
        else clear_bit(GPIO, sda);        
        delay_100us(1);
        set_bit(GPIO, scl);       
        delay_100us(1);
        clear_bit(GPIO, scl);       
        delay_100us(1);
    }
}

void i2c_send_data(char buffer)
{
    for (j=7;j>=0;j--)
    {
        if (test_bit(buffer,j)) set_bit(GPIO, sda);         
        else clear_bit(GPIO, sda);        
        delay_100us(1);
        set_bit(GPIO, scl);       
        delay_100us(1);
        clear_bit(GPIO, scl);       
        delay_100us(1);
    }
}

unsigned char i2c_read_data(void)
{
    unsigned char rd_buffer;
    rd_buffer = 0;
    set_bit(TRISIO, sda);             // Sets the line to input to release it
    delay_100us(1);
    for (j=7;j>=0;j--)
    {
        set_bit(GPIO, scl);       
        delay_100us(1);
        if (test_bit(GPIO,sda)) rd_buffer = rd_buffer + (1<<j);       
        clear_bit(GPIO, scl);       
        delay_100us(1);
    }
    clear_bit(TRISIO, sda);             // Sets the line to output to release it
    delay_100us(1);
    return rd_buffer;
}

void i2c_send_write_bit(void)
{
    clear_bit(GPIO, sda);             
    delay_100us(1);
    set_bit(GPIO, scl);             
    delay_100us(1);
    clear_bit(GPIO, scl);             
    delay_100us(1);
}

void i2c_send_read_bit(void)
{
    set_bit(GPIO, sda);             
    delay_100us(1);
    set_bit(GPIO, scl);             
    delay_100us(1);
    clear_bit(GPIO, scl);             
    delay_100us(1);
}

void i2c_read_ack(void)
{
    set_bit(TRISIO, sda);             // Sets the line to input to release it
    delay_100us(1);
    set_bit(GPIO, scl);             
    delay_100us(1);
    clear_bit(GPIO, scl);             
    delay_100us(1);                   
    clear_bit(TRISIO, sda);             // Sets the line to input to release it
    delay_ms(1);                   // long delay to clearly separate bytes
}

void i2c_send_ack(void)
{
    clear_bit(GPIO, sda);             
    delay_100us(1);
    set_bit(GPIO, scl);             
    delay_100us(1);
    clear_bit(GPIO, scl);             
    delay_ms(1);                   // long delay to clearly separate bytes
}

void i2c_send_nack(void)
{
    set_bit(GPIO, sda);             
    delay_100us(1);
    set_bit(GPIO, scl);             
    delay_100us(1);
    clear_bit(GPIO, scl);             
    delay_ms(1);                   // long delay to clearly separate bytes
}

void i2c_write_reg(unsigned char address,unsigned char reg,unsigned char data)
{
    i2c_send_start();
    i2c_send_addr(address);
    i2c_send_write_bit();
    i2c_read_ack();
    i2c_send_data(reg);
    i2c_read_ack();
    i2c_send_data(data);
    i2c_read_ack();
    i2c_send_stop();
}

unsigned char i2c_read_reg(unsigned char address,unsigned char reg)
{
    unsigned char data;

    i2c_send_start();
    i2c_send_addr(address);
    i2c_send_write_bit();
    i2c_read_ack();
    i2c_send_data(reg);
    i2c_read_ack();
    i2c_send_stop();
    i2c_send_start();
    i2c_send_addr(address);
    i2c_send_read_bit();
    i2c_read_ack();
    data = i2c_read_data();
    i2c_send_ack();     // devrait [etre nack]
    i2c_send_stop();
    return data;
}

void main(void) 
{  	// Initialization stage
    //Set CPU to 8MHz Here.
    // Initialization stage
    INTCON = 0b00000000; // disable interrupts
    IOC = 0b00000000; // disable GPIO interrupts on change
    PIE1 = 0b00000000; // disable various interrupts
    OPTION_REG = 0b00001111; // weak pullups global enable,
    WPU = 0b00110111; // individual weak pullups enabled
    OSCCON = 0b01110001; // select internal oscillator and set it to 8Mhz
    T1CON = 0b10110111; // enable timer1 and set prescaler div. by 1
    T2CON = 0b01111010; // disable timer2 and set prescaler div. by 16 and postscaler div. by 16
    CCP1CON = 0b00000101; // Set Capture mode, every rising edge
    ADCON0 = 0b00000000; // disable A/D converter
    ANSEL = 0b00000000; // Disable analog assignment to pins
    CMCON0 = 0b00010111; // Disable comparator, use internal reference, invert output.
    CMCON1 = 0b00000000; // comparator gating timer 1 asynchronously
    VRCON = 0b00100001; // Disable internal comparator Vref, use low range, set Vref to 0.15V
    WDTCON = 0b00010100; // Assign the right prescaler value to the watchdog timer. Keep WTD off for now.
    GPIO   = 0b00110000; // Set output pins initially before enabling them.
    TRISIO = 0b00101011; // GPIO first disabled, but assigned as follows:
									  //	5-sda line
									  //	4-scl line
									  //	3-Input only. Interrupt line (future)
									  //	2-Externally Locked Lock LED output
									  //	1-Available
									  //	0-Available
	delay_ms(250);					// Generates a startup delay of 0.25 second

//Si5351C setup starts here   
    i2c_write_reg(i2c_addr,3,0b11111111);
    // power down all 8 output drivers
    i2c_write_reg(i2c_addr,16,0x08);
    i2c_write_reg(i2c_addr,17,0x08);
    i2c_write_reg(i2c_addr,18,0x08);
    i2c_write_reg(i2c_addr,19,0x08);
    i2c_write_reg(i2c_addr,20,0x08);
    i2c_write_reg(i2c_addr,21,0x08);
    i2c_write_reg(i2c_addr,22,0x08);
    i2c_write_reg(i2c_addr,23,0x08);     
    //Write new configuration
    i2c_write_reg(i2c_addr,15,0b00000100);   // PLL Input Source
    i2c_write_reg(i2c_addr,9,0b11111111);   // OEB Pin Enable Control Mask
    i2c_write_reg(i2c_addr,16,0b00001100);   // CLK0 Control
    i2c_write_reg(i2c_addr,26,0);   // Multisynth NA Parameter MSNA_P3[15:8]
    i2c_write_reg(i2c_addr,27,1);   // Multisynth NA Parameter MSNA_P3[7:0]
    i2c_write_reg(i2c_addr,28,0);   // Multisynth NA Parameter MSNA_P1[17:16]
    i2c_write_reg(i2c_addr,29,0x26);   // Multisynth NA Parameter MSNA_P1[15:8]
    i2c_write_reg(i2c_addr,30,0x80);   // Multisynth NA Parameter MSNA_P1[7:0]
    i2c_write_reg(i2c_addr,31,0);   // Multisynth NA Parameters MSNA_P3[19:16] MSNA_P2[19:16]
    i2c_write_reg(i2c_addr,32,0);   // Multisynth NA Parameter MSNA_P2[15:18]
    i2c_write_reg(i2c_addr,33,0);   // Multisynth NA Parameter MSNA_P2[7:0]
    i2c_write_reg(i2c_addr,42,0);   // Multisynth0 Parameter MS0_P3[15:8]
    i2c_write_reg(i2c_addr,43,1);   // Multisynth0 Parameter MS0_P3[7:0]
    i2c_write_reg(i2c_addr,44,0b00000000);   // Multisynth0 Parameters R0, DIVBY4, MS0_P1[17:16]
    i2c_write_reg(i2c_addr,45,0x0D);   // Multisynth0 Parameter MS0_P1[15:8]
    i2c_write_reg(i2c_addr,46,0x00);   // Multisynth0 Parameter MS0_P1[7:0]
    i2c_write_reg(i2c_addr,47,0);   // Multisynth0 Parameters MS0_P3[19:16] MS0_P2[19:16]
    i2c_write_reg(i2c_addr,48,0);   // Multisynth0 Parameter MS0_P2[15:8]
    i2c_write_reg(i2c_addr,49,0);   // Multisynth0 Parameter MS0_P2[7:0]
    i2c_write_reg(i2c_addr,187,0b01000000);     // Fanout Enable of the XO for passthrough fallback
    i2c_write_reg(i2c_addr,177,0b00100000);   // PLL Reset Suggested to write 0xAC  

    // Turn on output 0
    i2c_write_reg(i2c_addr,3,0b11111110);   
    delay_ms(250);					// Generates a startup delay of 0.25 second
//Si5351C setup ends here   
      

    // Reference Input monitoring loop
    while(1)
    {    
        status = i2c_read_reg(i2c_addr,0);  //Read the status Register 
        if (((status & 0b00010000) != 0) && (status != old_status))
        {   // Loss Of Signal or Loss of Lock detected on External Reference input
            i2c_write_reg(i2c_addr,16,0b00000000);   // Switch to 27MHz crystal pass-through
            i2c_write_reg(i2c_addr,177,0b00100000);   // PLL Reset
            set_bit(GPIO,LED);          // Turn off LED
        }
        else if (status != old_status)
        {   // 10MHz External reference Signal Present
            i2c_write_reg(i2c_addr,16,0b00001100);   // Switch 10 MHz external reference
            i2c_write_reg(i2c_addr,177,0b00100000);   // PLL Reset
            clear_bit(GPIO,LED);        // Turn on LED
        }
        old_status = status;
    } 
}


/* The following parameters are used to convert a 10.000 MHz REFIN frequency to 28.800 MHz output frequency
 * 
 * 10 MHz x 81
 * -----------  = 28.8 MHz
 *   28.125
 * 
 * MSNB
 * ----
 * a=81, b=0, c=1  ->   MSNA_P1 = (128 x 81) - 512 = 9856 = 0x2680
 *                      MSNA_P2 = 0
 *                      MSNA_P3 = 1
 * MS1
 * ---
 * a=28, b=1, c=8  ->   MS0_P1 = (128 x 28) + (128 x 1/8)- 512 = 3088 = 0x0C10
 *                      MS0_P2 = 128 - (8 x 16) = 0
 *                      MS0_P3 = 8

 */ 
// This code would set up 27 MHz on the main output and 28.8MHz on the spare output. 28.8 MHz is typically used on RTL-SDR receivers.
//
//Si5351C setup starts here   
//    i2c_write_reg(i2c_addr,3,0b11111111);
//      // power down all 8 output drivers
//    i2c_write_reg(i2c_addr,16,0x08);
//    i2c_write_reg(i2c_addr,17,0x08);
//    i2c_write_reg(i2c_addr,18,0x08);
//    i2c_write_reg(i2c_addr,19,0x08);
//    i2c_write_reg(i2c_addr,20,0x08);
//    i2c_write_reg(i2c_addr,21,0x08);
//    i2c_write_reg(i2c_addr,22,0x08);
//    i2c_write_reg(i2c_addr,23,0x08);     
//      // Configure PLLA, CLK0 and CLK1 outputs
//    i2c_write_reg(i2c_addr,15,0b00001100);   // PLL A and B Input Source
//    i2c_write_reg(i2c_addr,9,0b11111111);   // OEB Pin Enable Control Mask
//    i2c_write_reg(i2c_addr,16,0b00001100);   // CLK0 Control, Drive strength
//    i2c_write_reg(i2c_addr,17,0b00101111);   // CLK1 Control, Drive strength
//      // Multisynth NA Settings to 810 MHz
//    i2c_write_reg(i2c_addr,26,0);   // Multisynth NA Parameter MSNA_P3[15:8]
//    i2c_write_reg(i2c_addr,27,1);   // Multisynth NA Parameter MSNA_P3[7:0]
//    i2c_write_reg(i2c_addr,28,0);   // Multisynth NA Parameter MSNA_P1[17:16]
//    i2c_write_reg(i2c_addr,29,0x26);   // Multisynth NA Parameter MSNA_P1[15:8]
//    i2c_write_reg(i2c_addr,30,0x80);   // Multisynth NA Parameter MSNA_P1[7:0]
//    i2c_write_reg(i2c_addr,31,0);   // Multisynth NA Parameters MSNA_P3[19:16] MSNA_P2[19:16]
//    i2c_write_reg(i2c_addr,32,0);   // Multisynth NA Parameter MSNA_P2[15:18]
//    i2c_write_reg(i2c_addr,33,0);   // Multisynth NA Parameter MSNA_P2[7:0]
//      // Multisynth NB Settings to 810 MHz
//    i2c_write_reg(i2c_addr,34,0);   // Multisynth NA Parameter MSNA_P3[15:8]
//    i2c_write_reg(i2c_addr,35,1);   // Multisynth NA Parameter MSNA_P3[7:0]
//    i2c_write_reg(i2c_addr,36,0);   // Multisynth NA Parameter MSNA_P1[17:16]
//    i2c_write_reg(i2c_addr,37,0x26);   // Multisynth NA Parameter MSNA_P1[15:8]
//    i2c_write_reg(i2c_addr,38,0x80);   // Multisynth NA Parameter MSNA_P1[7:0]
//    i2c_write_reg(i2c_addr,39,0);   // Multisynth NA Parameters MSNA_P3[19:16] MSNA_P2[19:16]
//    i2c_write_reg(i2c_addr,40,0);   // Multisynth NA Parameter MSNA_P2[15:18]
//    i2c_write_reg(i2c_addr,41,0);   // Multisynth NA Parameter MSNA_P2[7:0]
//      // Multisynth 0  Settings to 27.0 MHz
//    i2c_write_reg(i2c_addr,42,0);   // Multisynth0 Parameter MSNA_MS0_P3[15:8]
//    i2c_write_reg(i2c_addr,43,1);   // Multisynth0 Parameter MSNA_MS0_P3[7:0]
//    i2c_write_reg(i2c_addr,44,0b00000000);   // Multisynth0 Parameters R0, DIVBY4, MS0_P1[17:16]
//    i2c_write_reg(i2c_addr,45,0x0D);   // Multisynth0 Parameter MS0_P1[15:8]
//    i2c_write_reg(i2c_addr,46,0x00);   // Multisynth0 Parameter MS0_P1[7:0]
//    i2c_write_reg(i2c_addr,47,0);   // Multisynth0 Parameters MS0_P3[19:16] MS0_P2[19:16]
//    i2c_write_reg(i2c_addr,48,0);   // Multisynth0 Parameter MS0_P2[15:8]
//    i2c_write_reg(i2c_addr,49,0);   // Multisynth0 Parameter MS0_P2[7:0]
//      // Multisynth 1 Settings to 28.8 MHz
//    i2c_write_reg(i2c_addr,50,0);   // Multisynth1 Parameter MSNA_MS1_P3[15:8]
//    i2c_write_reg(i2c_addr,51,8);   // Multisynth1 Parameter MSNA_MS1_P3[7:0]  
//    i2c_write_reg(i2c_addr,52,0b00000000);   // Multisynth1 Parameters R0, DIVBY4, MS1_P1[17:16]
//    i2c_write_reg(i2c_addr,53,0x0C);   // Multisynth1 Parameter MS1_P1[15:8]
//    i2c_write_reg(i2c_addr,54,0x10);   // Multisynth1 Parameter MS1_P1[7:0]
//    i2c_write_reg(i2c_addr,55,0);   // Multisynth1 Parameters MS1_P3[19:16] MS1_P2[19:16]
//    i2c_write_reg(i2c_addr,56,0);   // Multisynth1 Parameter MS1_P2[15:8]
//    i2c_write_reg(i2c_addr,57,0);   // Multisynth1 Parameter MS1_P2[7:0]
//
//    i2c_write_reg(i2c_addr,187,0b01000000);     // Fanout Enable of the XO for passthrough fallback
//    i2c_write_reg(i2c_addr,177,0b10100000);   // PLLs Reset (Suggested to write 0xAC)
//
//// Turn on outputs 0 and 1
//    i2c_write_reg(i2c_addr,3,0b11111100);   
//
//    delay_ms(250);					// Generates a startup delay of 0.25 second
//Si5351C setup ends here   
